package org.msrb.emma.client;

import com.google.gson.Gson;
import com.google.gson.GsonBuilder;
import java.io.File;
import java.io.FileOutputStream;
import java.io.FileReader;
import java.io.StringReader;
import java.net.MalformedURLException;
import java.net.URL;
import java.text.ParseException;
import java.text.SimpleDateFormat;
import java.util.Date;
import java.util.TimeZone;
import java.util.List;
import java.util.PropertyResourceBundle;
import java.util.ResourceBundle;
import javax.xml.bind.JAXBContext;
import javax.xml.bind.JAXBException;
import javax.xml.bind.Unmarshaller;
import org.apache.commons.cli.BasicParser;
import org.apache.commons.cli.CommandLine;
import org.apache.commons.cli.HelpFormatter;
import org.apache.commons.cli.Options;
import org.msrb.emma.client.domain.continuingdisclosure.ContinuingDisclosureSubmissionDetailsType;
import org.msrb.emma.client.domain.financialreporting.FinancialReportingSubmissionDetailsType;
import org.msrb.emma.client.domain.primarymarket.SubmissionDetailsType;
import org.msrb.emma.client.domain.shortdocs.ShortSubmissionDetailsType;
import org.msrb.emma.subscriber.EmmaSubscriberCertificateClient;
import org.msrb.emma.subscriber.EmmaSubscriberClient;
import org.msrb.emma.subscriber.EmmaSubscriberPasswordClient;
import org.msrb.emma.subscriber.EmmaTransaction;
import org.msrb.emma.subscriber.EmmaTransactionList;
import org.msrb.emma.subscriber.SingleEmmaTransaction;

/**
 *
 * @author admin
 */
public class SubscriberClient {

    private static final String SubmissionPostingDateTimeOption = "spd";
    private static final String SubmissionIdentifierOption = "sid";
    private static final String TransactionListOption = "list";
    private static final String SubmissionTypeOption = "type";
    private static final String SavePathOption = "path";
    private static final String FileIdentifierOption = "fid";
    private static final String UserNameOption = "user";
    private static final String PasswordOption = "password";
    private static final String ServiceLocationOption = "url";
    private static final String CertificateAuthenticationOption = "cert";
    private static final String JsonOutputOption = "json";
    private static final String SpecVersionOption = "sver";

    private static final SimpleDateFormat SoapDateTimeFormatter = new SimpleDateFormat("yyyy-MM-dd'T'HH:mm:ss");
    private static final SimpleDateFormat DateTimeFormatter = new SimpleDateFormat("MM/dd/yyyy HH:mm:ss");
    private static final SimpleDateFormat DateOnlyFormatter = new SimpleDateFormat("MM/dd/yyyy");

    private static EmmaSubscriberClient client;
    private static ResourceBundle properties;
    private static CommandLine switches;

    public static void main(String[] args) throws Exception {
        DateTimeFormatter.setTimeZone(TimeZone.getTimeZone("America/New_York"));
        DateOnlyFormatter.setTimeZone(TimeZone.getTimeZone("America/New_York"));
        SoapDateTimeFormatter.setTimeZone(TimeZone.getTimeZone("America/New_York"));
        
        switches = getCommandLineSwitches(args);
        if (!isValid(switches)) {
            new HelpFormatter().printHelp("java org.msrb.emma.client.SubscriberClient [OPTIONS]", getOptions());
            return;
        }
        properties = new PropertyResourceBundle(new FileReader("connection.properties"));
        client = getClient();
        if (switches.hasOption(TransactionListOption)) {
            displayTransactionList(switches.getOptionValue(TransactionListOption), switches.getOptionValue(SubmissionTypeOption), switches.getOptionValue(SpecVersionOption));
        }
        if (switches.hasOption(SubmissionIdentifierOption)) {
            displaySingleTransaction(switches.getOptionValue(SubmissionIdentifierOption), switches.getOptionValue(SubmissionTypeOption), switches.getOptionValue(SubmissionPostingDateTimeOption), switches.hasOption(JsonOutputOption), switches.getOptionValue(SpecVersionOption));
        }
        if (switches.hasOption(FileIdentifierOption)) {
            retrieveFile(switches.getOptionValue(FileIdentifierOption), switches.getOptionValue(SavePathOption));
        }
    }

    private static EmmaSubscriberClient getClient() throws MalformedURLException {
        if (switches.hasOption(CertificateAuthenticationOption)) {
            return new EmmaSubscriberCertificateClient(getServiceUrl());
        }
        return new EmmaSubscriberPasswordClient(getServiceUrl(), getUserName(), getPassword());
    }

    private static CommandLine getCommandLineSwitches(String[] args) {
        try {
            return new BasicParser().parse(getOptions(), args);
        } catch (org.apache.commons.cli.ParseException ex) {
            return null;
        }
    }

    private static String getPassword() {
        if (switches.hasOption(PasswordOption)) {
            return switches.getOptionValue(PasswordOption);
        } else if (properties.containsKey(PasswordOption)) {
            return properties.getString(PasswordOption);
        } else {
            throw new IllegalStateException("Password not specified in connection.properties or command line.");
        }
    }

    private static String getUserName() {
        if (switches.hasOption(UserNameOption)) {
            return switches.getOptionValue(UserNameOption);
        } else if (properties.containsKey(UserNameOption)) {
            return properties.getString(UserNameOption);
        } else {
            throw new IllegalStateException("User name not specified in connection.properties or command line.");
        }
    }

    private static URL getServiceUrl() throws MalformedURLException {
        String address;
        if (switches.hasOption(ServiceLocationOption)) {
            address = switches.getOptionValue(ServiceLocationOption);
        } else if (properties.containsKey(ServiceLocationOption)) {
            address = properties.getString(ServiceLocationOption);
        } else {
            throw new IllegalStateException("Service location not specified in connection.properties or command line.");
        }
        return new URL(address);
    }

    private static void displayTransactionList(String requestDate, String submissionType, String specVersion) throws Exception {
        final EmmaTransactionList response = client.getTransactions(submissionType, getParsedDate(requestDate), specVersion);
        if (response.getStatusCode() != 200) {
            System.out.println("Got response code: " + response.getStatusCode());
            return;
        }
        final List<EmmaTransaction> transactions = response.getTransactions().getValue().getEmmaTransaction();
        System.out.println(String.format("Got %s transactions.", transactions.size()));
        for (EmmaTransaction transaction : transactions) {
            System.out.println(String.format("SubmissionID=%s, Posted=%s", transaction.getSubmissionId().getValue(), transaction.getPostingDateTime()));
        }
    }

    private static Date getParsedDate(String requestDate) throws ParseException {
        if (requestDate.contains("T")) {
            return SoapDateTimeFormatter.parse(requestDate);
        }
        return requestDate.contains(":") ? DateTimeFormatter.parse(requestDate) : DateOnlyFormatter.parse(requestDate);        
    }

    private static void displaySingleTransaction(String submissionId, String submissionType, String postedDateTime, boolean jsonFormat, String specVersion) throws Exception {
        final SingleEmmaTransaction details = client.getTransaction(submissionType, getParsedDate(postedDateTime), submissionId, specVersion);
        if (details.getStatusCode() != 200) {
            System.out.println("Got status code: " + details.getStatusCode());
            return;
        }
        final String xml = details.getReturnData().getValue();
        if (xml == null) {
            System.out.println("Transaction not found.");
        } else if (jsonFormat) {
            System.out.println("Single transaction=\r\n" + convertToJson(getDeserializedSubmission(submissionType, xml)));
        } else {
            System.out.println("Single transaction=\r\n" + xml);
        }
    }

    private static void retrieveFile(String fileId, String outputPath) throws Exception {
        try {
            final byte[] buffer = client.getFile(fileId);
            new FileOutputStream(new File(outputPath)).write(buffer);
        } catch (org.msrb.emma.subscriber.IEmmaSubscriptionServiceGetFileFileRetrievalFaultFaultFaultMessage ex) {
            System.out.println("Got status code: " + ex.getFaultInfo().getStatusCode());
        }
    }

    private static Options getOptions() {
        final Options options = new Options();
        options.addOption(TransactionListOption, true, "The date for which a list of transactions will be requested.  Must be used with -type.");
        options.addOption(SubmissionIdentifierOption, true, "The submission identifier for a transaction that will be requested.  Must be used with -spd and -type.");
        options.addOption(SubmissionPostingDateTimeOption, true, "The posting date/time for a transaction that will be requested.  Must be used with -sid and -type.");
        options.addOption(SubmissionTypeOption, true, "The type of submissions or transaction to be requested.  Must be used with either -sid or -list.");
        options.addOption(FileIdentifierOption, true, "The file identifier for a document to be requested.  Must be used with -path.");
        options.addOption(SavePathOption, true, "The full path of the file to save the requested document to.  Must be used with -fid.");
        options.addOption(UserNameOption, true, "Optional user name to authenticate with.  Overrides value in connection.properties if present.");
        options.addOption(PasswordOption, true, "Optional password to authenticate with.  Overrides value in connection.properties if present.");
        options.addOption(ServiceLocationOption, true, "Optional location of the WSDL file of the Subscriber Service to connect to.  Overrides value in connection.properties if present.");
        options.addOption(CertificateAuthenticationOption, false, "Flag indicating that the certificate specified in the WSIT file should be used for authentication instead of any specified user name/password combination.");
        options.addOption(JsonOutputOption, false, "Flag indicating that the results of a single transaction request should be output in JSON format");
        options.addOption(SpecVersionOption, true, "Spec version of data to return");
        return options;
    }

    private static boolean isValid(CommandLine options) {
        if (options == null) {
            return false;
        }
        boolean commandSpecified = false;
        if (options.hasOption(TransactionListOption)) {
            if (!options.hasOption(SubmissionTypeOption)) {
                return false;
            }
            commandSpecified = true;
        }
        if (options.hasOption(SubmissionIdentifierOption)) {
            if (!options.hasOption(SubmissionTypeOption)) {
                return false;
            } else if (!options.hasOption(SubmissionPostingDateTimeOption)) {
                return false;
            }
            commandSpecified = true;
        }
        if (options.hasOption(FileIdentifierOption)) {
            if (!options.hasOption(SavePathOption)) {
                return false;
            }
            commandSpecified = true;
        }
        return commandSpecified;
    }

    private static Object getDeserializedSubmission(final String submissionType, final String xml) throws JAXBException {
        return createSubmissionWrapper(submissionType, getUnmarshaller(submissionType).unmarshal(new StringReader(xml)));
    }

    private static Object createSubmissionWrapper(final String submissionType, final Object submission) {
        switch (submissionType) {
            case "CD":
                return new ContinuingDisclosureWrapper(submission);
            case "FR":
                return new FinancialReportingWrapper(submission);
            default:
                return new GenericSubmissionWrapper(submission);
        }
    }

    private static Unmarshaller getUnmarshaller(final String submissionType) throws JAXBException {
        JAXBContext context = null;
        switch (submissionType) {
            case "AV":
                context = JAXBContext.newInstance(ShortSubmissionDetailsType.class);
                break;
            case "CD":
                context = JAXBContext.newInstance(ContinuingDisclosureSubmissionDetailsType.class);
                break;
            case "FR":
                context = JAXBContext.newInstance(FinancialReportingSubmissionDetailsType.class);
                break;
            default:
                context = JAXBContext.newInstance(SubmissionDetailsType.class);
        }
        return context.createUnmarshaller();
    }

    private static String convertToJson(Object submission) {
        Gson converter = new GsonBuilder()
                .setPrettyPrinting()
                .create();
        return converter.toJson(submission);
    }

    private static class ContinuingDisclosureWrapper {

        private final Object submission;

        public ContinuingDisclosureWrapper(Object submission) {
            this.submission = submission;
        }

        public Object getContinuingDisclosureSubmission() {
            return submission;
        }
    }

    private static class GenericSubmissionWrapper {

        private final Object submission;

        public GenericSubmissionWrapper(Object submission) {
            this.submission = submission;
        }

        public Object getSubmission() {
            return submission;
        }
    }

    private static class FinancialReportingWrapper {

        private final Object submission;

        public FinancialReportingWrapper(Object submission) {
            this.submission = submission;
        }

        public Object getFinancialReportingSubmission() {
            return submission;
        }
    }

}
